# TODO: __all__ = []


from collections import namedtuple
import logging
import struct

from cffi import FFI

from .sdllibs import sdl_lib
# from .sdl import ffi as sdl_ffi
# from . import sdlconstants
from . import sdlpixels
from .rect import Rect


log = logging.getLogger(__name__)

ffi = FFI()


PixelFormat = namedtuple('PixelFormat', 'format palette bitsperpixel bytesperpixel' +
                         ' rmask gmask bmask amask rloss gloss bloss aloss rshift gshift bshift ashift refcount next')
PixelPalette = namedtuple('PixelPalette', 'ncolors color')
Color = namedtuple('Color', 'r g b a')


def pixel_format(cdata):
    """returns a PixelFormat
    """
    # TODO: this will be different for 8-bit surfaces with a palette
    buf = ffi.buffer(cdata)
    pixel_format_elements = struct.unpack('IIBBIIIIBBBBBBBBhI', buf)
    pixel_format = PixelFormat(*pixel_format_elements)
    # print(pixel_format)
    # print('format={}'.format(sdl.getPixelFormatName(pixel_format.format)))
    return pixel_format


class Surface(object):

    def __init__(self, size, flags=0, depth=0, masks=None, surface=None):
        width, height = size
        if depth == 0:
            depth = 32
        if surface is None:
            if not masks:
                masks = [0] * 4
            self.__sdl_surface = sdl_lib.SDL_CreateRGBSurface(flags, width, height, depth, *masks)
        elif isinstance(surface, Surface):
            # TODO: copy, format, blit the surface?
            # SDL_CreateRGBSurfaceFrom
            # http://wiki.libsdl.org/SDL_CreateRGBSurfaceFrom?highlight=%28\bCategoryAPI\b%29|%28SDLFunctionTemplate%29
            self.__sdl_surface = Surface(surface.get_size())
        else:
            self.__sdl_surface = surface
        self.__size = width, height
        self.__flags = flags
        self.__depth = depth
        self.__masks = masks

    def get_size(self):
        return self.__size

    def get_width(self):
        return self.__size[0]

    def get_height(self):
        return self.__size[1]

    def get_flags(self):
        return self.__flags

    def get_bitsize(self):
        return self.__depth

    def get_rect(self, **kwargs):
        w, h = self.__size
        r = Rect(0, 0, w, h)
        for k, v in kwargs.items():
            setattr(r, k, v)
        return r

    def fill(self, color, rect=None, special_flags=0):
        surface = self.__sdl_surface
        map_color = sdl_lib.SDL_MapRGBA if len(color) == 4 else sdl_lib.SDL_MapRGB

        # TODO: how to MUSTLOCK(surface)?
        # surface.lockSurface()
        if rect is None:
            size = self.get_size()
            rect = Rect(0, 0, size[0], size[1])
        elif not isinstance(rect, Rect):
            rect = Rect(*rect)
        # surface.fillRect(rect, map_color(surface.format, *color))
        sr = rect.sdl_rect
        sdl_lib.SDL_FillRect(surface, rect.sdl_rect, map_color(surface.format, *color))
        # TODO: how to MUSTLOCK(surface)?
        # surface.unlockSurface()
        return Rect(rect)

    def blit(self, source, dest_rect, area=None, special_flags=0):
        dest_surface = self.__sdl_surface
        # TODO: how to MUSTLOCK(surface)?
        # dest_surface.lockSurface()
        if area is None:
            size = source.get_size()
            area = Rect(0, 0, int(size[0]), int(size[1]))
        elif not isinstance(area, Rect):
            area = Rect(*area)
        if not isinstance(dest_rect, Rect):
            dest_rect = Rect(dest_rect)
        sdl_lib.SDL_UpperBlit(source.sdl_surface, area.sdl_rect, dest_surface, dest_rect.sdl_rect)
        # TODO: how to MUSTLOCK(surface)?
        # dest_surface.unlockSurface()

    def blit_scaled(self, source, dest_rect, area=None):
        # def round_rect(r):
        #     x, y, w, h = r[0:4]
        #     x = int(round(x))
        #     w = int(round(w + r[0] - x))
        #     y = int(round(y))
        #     h = int(round(h + r[1] - y))
        #     return x, y, w, h
        dest_surface = self.__sdl_surface
        if area is None:
            # size = source.get_size()
            # area = Rect(0, 0, int(size[0]), int(size[1]))
            area = source.get_rect()
        elif not isinstance(area, Rect):
            area = Rect(*area)
        if not isinstance(dest_rect, Rect):
            dest_rect = Rect(dest_rect)
        sdl_lib.SDL_UpperBlitScaled(source.sdl_surface, area.sdl_rect, dest_surface, dest_rect.sdl_rect)

    def convert(self, format=None):
        surf = get_window_list()[0].surface.sdl_surface
        if format is None:
            format = surf.format
        new_surf = self.__sdl_surface.convertSurface(format, 0)
        if new_surf is None:
            # TODO: proper exception
            raise Exception('could not convert surface')
        return Surface((new_surf.w, new_surf.h), surface=new_surf)

    def convert_alpha(self, format=None):
        # This gets the best blit performance on my system.
        # new format=SDL_PIXELFORMAT_ARGB8888
        # window format=SDL_PIXELFORMAT_RGB888
        # surface1 format=SDL_PIXELFORMAT_ABGR8888
        # surface2 format=SDL_PIXELFORMAT_ARGB8888  <<<
        converted_surface = self
        if format:
            surf = sdl_lib.SDL_ConvertSurfaceFormat(self.sdl_surface, format, 0)
            converted_surface = Surface((surf.w, surf.h), surface=surf)
        else:
            # TODO: there's probably a more elegant way to do this
            window_format = pixel_format(get_window_list()[0].surface.sdl_surface.format).format
            surface_format = pixel_format(self.sdl_surface.format).format
            surface_layout = sdlpixels.layout_to_name[sdlpixels.pixel_layout(surface_format)]
            window_order = sdlpixels.order_to_name[sdlpixels.pixel_order(window_format)]

            if sdlpixels.is_pixel_format_alpha(surface_format):
                target_order = window_order.replace('X', 'A')
                format_name = 'SDL_PIXELFORMAT_' + target_order + surface_layout
                log.debug('convert_alpha: new format={}'.format(format_name))
                format = getattr(sdlpixels, format_name)
                surf = sdl_lib.SDL_ConvertSurfaceFormat(self.sdl_surface, format, 0)
                converted_surface = Surface((surf.w, surf.h), surface=surf)
            elif sdlpixels.is_pixel_format_indexed(surface_format):
                # TODO
                pass
            elif sdlpixels.is_pixel_format_fourcc(surface_format):
                # TODO
                pass
        return converted_surface

    def __getsdlsurface(self):
        return self.__sdl_surface
    sdl_surface = property(__getsdlsurface)

    def __del__(self):
        # TODO: unreliable
        if self.__sdl_surface:
            try:
                garbage = self.__sdl_surface
                self.__sdl_surface = None
                sdl_lib.SDL_FreeSurface(garbage)
            except Exception as e:
                log.info('error: sdl_lib.SDL_FreeSurface() failed')
                log.info(e)


from .window import get_list as get_window_list

"""
dir(surface):
cdata
clip_rect
convertSurface
convertSurfaceFormat
createColorCursor
createSoftwareRenderer
fillRect
fillRects
flags
format
freeSurface
getClipRect
getColorKey
getSurfaceAlphaMod
getSurfaceBlendMode
getSurfaceColorMod
h
lockSurface
lock_data
locked
lowerBlit
lowerBlitScaled
map
pitch
pixels
refcount
saveBMP_RW
setClipRect
setColorKey
setSurfaceAlphaMod
setSurfaceBlendMode
setSurfaceColorMod
setSurfacePalette
setSurfaceRLE
softStretch
unlockSurface
upperBlit
upperBlitScaled
userdata
w
"""
